""" Orange data table and other data manipulation utilities """

# Data hints variables
TAX_ID = 'Taxonomy_id'

# Will be set to True if gene names are represented as attribute names.
# If gene names are in rows, we set this value to False. (user must select proper column index)
GENE_NAME = 'gene_as_attribute_name'
