DOMAIN = 'gene_sets'

# cytoband
CYTOBAND_DOWNLOAD_LINK = 'http://statweb.stanford.edu/~tibs/GSA/cytobands-stanford.gmt'

# Reactome
REACTOME_DOWNLOAD_LINK = 'http://www.reactome.org/download/current/ReactomePathways.gmt.zip'
REACTOME_FILE_NAME = 'ReactomePathways.gmt'

# GO
GO_TERM_LINK = 'http://amigo.geneontology.org/cgi-bin/amigo/term-details.cgi?term={}'

# OMIM
OMIM_LINK = 'http://www.omim.org/entry/{}'
